require 'curses'

# dnf install ruby-tcltk tkimg

#===============================================================================
#
#	Class: LegacyGraphics
#
class LegacyGraphics < Curses::Window

	$curses_enable = true

	def initialize(height, width, y, x, gy = 400, gx = 640, cy = 16, cx = 8, c = nil, out = {})
		@x = width
		@y = height
		@gx = gx
		@gy = gy
		@cx = cx
		@cy = cy
		@lx = 0
		@ly = 0
		super(height, width, y, x)
	end

	#---------------------------------------------------------------------------
	#
	#	pset
	#
	def pset(x, y, c)
		_pset(x * @x / @gx, y * @y / @gy, c)
	end

	#---------------------------------------------------------------------------
	#
	#	_pset
	#
	def _pset(x, y, c)
		self.setpos(y, x)
		self.addch(_color(c))
	end

	#---------------------------------------------------------------------------
	#
	#	locate
	#
	def locate(x, y)
		self.setpos(y * @cy * @y / @gy, x * @cx * @x / @gx)
	end

	#---------------------------------------------------------------------------
	#
	#	print
	#
	def print(mes)
		self.addstr(mes)
	end

	#---------------------------------------------------------------------------
	#
	#	line
	#
	def line(x1, y1, x2, y2, c, bf = '', w = 1, d = nil, a = nil)
			  cx1 = (x1 * @x / @gx).to_i
			  cy1 = (y1 * @y / @gy).to_i
		@lx = cx2 = (x2 * @x / @gx).to_i
		@ly = cy2 = (y2 * @y / @gy).to_i
		self._line(cx1, cy1, cx2, cy2, c)
	end

	#---------------------------------------------------------------------------
	#
	#	cline
	#
	def cline(x, y, c)
			  cx1 = @lx
			  cy1 = @ly
		@lx = cx2 = (x * @x / @gx).to_i
		@ly = cy2 = (y * @y / @gy).to_i
		self._line(cx1, cy1, cx2, cy2, c)
	end

	#---------------------------------------------------------------------------
	#
	#	_line
	#
	def _line(cx1, cy1, cx2, cy2, c, ras = true)
		if(cx1 == cx2 and cy1 == cy2)							# POINT
			self.setpos(cy1, cx1)
			self.addch(c.class <= Numeric ? '.1234567'[c] : c)
		elsif((cx1 - cx2).abs > (cy1 - cy2).abs)
			([cx1, cx2].min..[cx1, cx2].max).each {|x|			# YOKO
				prc = (x - cx1).to_f / (cx2 - cx1)
				y = (cy2 - cy1) * prc + cy1
				self.setpos(y, x)
				self.addch(c.class <= Numeric ? '.1234567'[c] : c)
			}
		else
			([cy1, cy2].min..[cy1, cy2].max).each {|y|			# TATE
				prc = (y - cy1).to_f / (cy2 - cy1)
				x = (cx2 - cx1) * prc + cx1
				self.setpos(y, x)
				self.addch(c.class <= Numeric ? '.1234567'[c] : c)
			}
		end
		@ras and ras and ([cy1, cy2].min..[cy1, cy2].max).each {|y|	# RASTERIZE
			@ras[y] ||= []
			if(cy2 - cy1 != 0)
				prc = (y - cy1).to_f / (cy2 - cy1)
				x = (cx2 - cx1) * prc + cx1
				@ras[y] << x.to_i
			else
				@ras[y] << cx1 << cx2
			end
		}
	end

	#---------------------------------------------------------------------------
	#
	#	circle	(http://en.wikipedia.org/wiki/Midpoint_circle_algorithm)
	#
	def circle(x, y, r, c, f = '', dy = 1, opt = {})
		cx = x * @x / @gx; cy = y * @y / @gy
		x0 = r * @y / @gy; y0 = 0; re = 1 - x0; while(x0 >= y0)
			_fill(cx - y0, it = cy - x0 / dy, cx + y0, it, c, f1 = (f =~ /f/i))	#  / \ upper
			_fill(cx - x0, it = cy - y0 / dy, cx + x0, it, c, f1)				# /   \
			_fill(cx - x0, it = cy + y0 / dy, cx + x0, it, c, f1)				# \   /
			_fill(cx - y0, it = cy + x0 / dy, cx + y0, it, c, f1)				#  \ / lower
			re += 2 * ((y0 += 1) - (re < 0 ? 0 : x0 -= 1) + 1)
		end
	end

	def fill(x1, y1, x2, y2, c)
		(y1..y2).each {|y|
			line(x1, y, x2, y, c)
		}
	end

	def _fill(x1, y1, x2, y2, c, f)
		f ? _line(x1, y1, x2, y2, c) : (_pset(x1, y1, c); _pset(x2, y2, c))
	end

	def polygon(points, c, f = '')				# points = [[x1, y1], [x2, y2], [x3, y3]...]
		f =~ /f/i and @ras = {}
		ps = points.dup; xy0 = ps.shift
		line(*xy0, *(xy = ps.shift), c)
		cline(*xy, c) while(xy = ps.shift)
		cline(*xy0, c)
		@ras and @ras.sort.each {|y, xs|
			xs.sort!; x1 = x2 = nil
			_line(x1, y, x2, y, c, false) while(x1 = xs.shift and x2 = xs.shift)
		}
		@ras = nil
	end

	def put_image_png(x, y, file)
		text(x, y, file, nil)
	end

	def _text(x, y, s, c, ps = {})
		ps[:get_size] and return([s.size * @cx, @cy])
		locate(x / @cx, y / @cy)
		print(s)
	end

	def _text_bmf(x, y, s, c, ps = {})
		tx = ps[:tx] || 1; ty = ps[:ty] || 2; raster = ps[:raster] || 0 # depend :ty
		cx = 0; cy = 0; s.each_byte {|code|
			if(code < 0x20)
				code == 0x0a and (cx = 0; cy += 1; next)
				code =  0x20
			end
 			code > 0x7F and code = 0x23	# '#'
			_text(x + cx * 8 * tx, y + cy * 8 * ty, code.chr, c)
			cx += 1
		}
	end

	def symbol(x, y, s, tx, ty, c, ps = {})
		_text_bmf(x, y, s, c, ps = { :tx => tx, :ty => ty })
	end

	def rotate(r = nil, x = 0, y = 0)
		# not implement yet
		yield
	end

	def _color(c, color = '.1234567')
		c.is_a?(Numeric) and return(color[c])
		return('8')
	end

	def refresh(fps = nil)
		super()
	end

	def method_missing(func, *args)
		func =~ /3d$/ and return
		raise
	end
end

