# coding: utf-8

require './TrueLegacyGraphicsCommon'
require 'gtk2'
begin
	require 'rmagick'
rescue ScriptError
	require 'RMagick'
end

#===============================================================================
#
#	Class: LegacyGraphics
#
class LegacyGraphics < TrueLegacyGraphicsCommon

	@@color = []
	pow = { '0' => 0, '1' => 65535 }; (0..7).each {|c|		# イロイッカイズツ
		grb = '%03b' % c; @@color << Gdk::Color.new(pow[grb[1]], pow[grb[0]], pow[grb[2]])
	}
	@@colormap = Gdk::Colormap.system
	@@color.each {|c| @@colormap.alloc_color(c, false, true) }

	def self.get_image_size_png(file)
		@@png_image = Magick::Image.read(@@cur_png = file)
		[@@png_image[0].columns, @@png_image[0].rows]
	end

	def self.point(x, y)
		@@png_image[0].pixel_color(x, y)
	end

	def initialize(h_dmy, w_dmy, y_dmy, x_dmy, gy = 400, gx = 640, cy = 16, cx = 8, c = 0, out = {})
		@lx = 0
		@ly = 0

		@window = Gtk::Window.new
		@window.signal_connect('destroy') { Gtk.main_quit }
		@window.title = 'TrueLegacyGraphicsGtk2'
		@window.set_size_request(gx, gy)
		@window.set_app_paintable(true)
		@window.realize

		# http://ruby-gnome2.sourceforge.jp/hiki.cgi?Gdk%3A%3ADrawable
		@drawable = @window.window
		@gc = Gdk::GC.new(@drawable)

		@draw_procs = []
		fill(0, 0, gx - 1, gy - 1, c)
	end

	#---------------------------------------------------------------------------
	#
	#	_line
	#
	def _line(x1, y1, x2, y2, c, w, d, a)

		@draw_procs << Proc.new {
			@gc.set_foreground(@@color[c.class <= Numeric ? c : 7])
			@drawable.draw_line(@gc, x1, y1, x2, y2)
		}
		pset(x2, y2, c)
	end

	def circle(x, y, r, c, f = '', dy = 1, opt = {})
		# not yet
	end

	def fill(x1, y1, x2, y2, c)

		@draw_procs << Proc.new {
			@gc.set_foreground(@@color[c.class <= Numeric ? c : 7])
			@drawable.draw_rectangle(@gc, true, (x1 < x2 ? x1 : x2), (y1 < y2 ? y1 : y2), (x1 - x2).abs + 1, (y1 - y2).abs + 1)
		}
	end

	def polygon(points, c, f = '', w = 1)		# points = [[x1, y1], [x2, y2], [x3, y3]...]

		@draw_procs << Proc.new {
			@gc.set_foreground(@@color[c.class <= Numeric ? c : 7])
			@drawable.draw_polygon(@gc, f =~ /f/i, points)
		}
	end

	def put_image_png(x, y, file)

		File.exist?(file) or raise('File not found.[%s]' % file)
		@draw_procs << Proc.new {
			@drawable.draw_pixbuf(@gc, Gdk::Pixbuf.new(file), 0, 0, x, y, -1, -1, Gdk::RGB::DITHER_NONE, 0, 0)
		}
	end

	def _text(x, y, s, c, ps = {})
#		{ 'font.family' => 'Helvetica', 'font.weight' => 400, 'font.size' => 10, 'layout.width' => 128, 'layout.alignment' => 'left', 'get_size' => true }

		# http://ruby-gnome2.sourceforge.jp/hiki.cgi?Pango%3A%3AFontDescription
		fontdesc = Pango::FontDescription.new
		fontdesc.family = ps[:font_family] || 'Helvetica'
		fontdesc.weight = ps[:font_weight] || Pango::WEIGHT_NORMAL # WEIGHT_NORMAL: 400, WEIGHT_BOLD: 700
		fontdesc.absolute_size = (ps[:font_size] || 10) * Pango::SCALE

		# http://ruby-gnome2.sourceforge.jp/hiki.cgi?Pango%3A%3ALayout
		layout = Pango::Layout.new(@window.pango_context)
		layout.font_description = fontdesc

		layout.width = (ps[:layout_width] || 128) * Pango::SCALE	# x ------- width(dot) ->|
		layout.alignment = Pango::Layout::ALIGN_LEFT				# |left....center...right|
		layout.alignment = Pango::Layout::ALIGN_CENTER	if(it = ps[:layout_alignment] and it =~ /center/i)
		layout.alignment = Pango::Layout::ALIGN_RIGHT	if(it = ps[:layout_alignment] and it =~ /right/i)
		layout.text = s

		ps[:get_size] and return(layout.pixel_size)
		@draw_procs << Proc.new {
			@gc.set_foreground(@@color[c.class <= Numeric ? c : 7])
			@drawable.draw_layout(@gc, x, y, layout)
		}
	end

	def refresh(fps = nil)
		@window.signal_connect('expose_event') {
			@draw_procs.each {|proc| proc.call }
		} unless(fps)
		@window.show_all

		#-------------------------------------------------------
		#
		#	垂直帰線割り込みのシミュレート
		#
		$jiffies = 0
		fps and @spf = (1.0 / fps)								# フレーム速度(1/FPS)
		Gtk.idle_add {
			$jiffies += 1
			@start = Time.now.to_f; @next = @start + @spf

			proc = @draw_procs.shift and proc.call
			@finish = Time.now.to_f
#			puts 'now: %13.2f, load: %6.2f%%' % [@start, (@finish - @start) * 100 / @spf]
			(it = @next - @finish) > 0 and sleep(it)			# 余った時間分 sleep する
			proc
		} if(fps)

		Gtk.main
	end
end

