# coding: utf-8

require './TrueLegacyGraphicsCommon'

require 'opengl'
require 'gl';	include GL
require 'glu';	include GLU
require 'glut';	include GLUT

begin
	require 'rmagick'
rescue ScriptError
	require 'RMagick'
end

include Math

def rad(d)
	d * PI / 180.0
end

def deg(r)
	r * 180 / PI
end

class PG < Array

	def p2v(p1, p2)
		[	p2[1] - p1[1],
			p2[2] - p1[2],
			p2[3] - p1[3]	]
	end

	def v2nv(v1, v2)
		[	v1[1] * v2[2] - v1[2] * v2[1],
			v1[2] * v2[0] - v1[0] * v2[2],
			v1[0] * v2[1] - v1[1] * v2[0]	]
	end

	def begin
		self.clear
	end

	def color(r, g, b)
		self.push([:COLOR, r, g, b])
	end

	def material(r, g, b)
		self.push([:MATERIAL, r, g, b, 1.0])
	end

	def texture(texture)
		self.push([:TEXTURE, texture[:W], texture[:H], texture[:PIXELS]])
	end

	def texcoord(x, y)
		self.push([:TEXCOORD, x, y])
	end

	def point(x, y, z)											# 反時計回りで指定
		self.push([:POINT, x, y, z])
	end

	def end(type = GL_TRIANGLE_FAN)
		enable_texture = false; self.each {|cmd|
			if(cmd[0] == :TEXTURE)
				glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, cmd[1], cmd[2], 0, GL_RGBA, GL_UNSIGNED_BYTE, cmd[3])
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
				glEnable(GL_TEXTURE_2D); enable_texture = true; #p 'en TX'
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
				glEnable(GL_BLEND); #p 'en BL'
			end
		}
		glBegin(type)
		ps = []; self.each {|cmd|
			if(cmd[0] == :POINT)								# 法線ベクトルの自動計算
				ps << cmd
				if(ps.size == 3)
					v1 = p2v(ps[0], ps[1])
					v2 = p2v(ps[1], ps[2])
					nv = v2nv(v1, v2)
					glNormal3f(*nv)
					break
				end
			end
		}
		self.each {|cmd|
			if(cmd[0] == :COLOR)
				glColor3f(cmd[1], cmd[2], cmd[3], 1.0)

			elsif(cmd[0] == :MATERIAL)
				glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, cmd[1, 3])

			elsif(cmd[0] == :POINT)
				glVertex3f(cmd[1], cmd[2], cmd[3])

			elsif(cmd[0] == :BINDTEX)
#				glBindTexture(GL_TEXTURE_2D, (GLuint) pg_p[i].v[0])

			elsif(cmd[0] == :TEXCOORD)
				glTexCoord2f(cmd[1], cmd[2])
			end
		}
		glEnd
		enable_texture and glDisable(GL_TEXTURE_2D); #p 'de TX'
		enable_texture and glDisable(GL_BLEND); #p 'de BL'
	end
end
#glEnable(GL_BLEND)
#glDisable(GL_BLEND)

#-----------------------------------------------
#
#	形状基底
#
class Model0

	# 原点を中心に任意の大きさの面を出現させる
	def self.surface(w, h, c)									# 幅, 高さ, 色
	  if(c.class == Array)
		pg = PG.new
		glScale(      w,      h,    1.0)
		pg.begin												# 前面
		pg.material(*c)
		pg.point(   0.5,    0.5,    0.0)
		pg.point(   0.5,   -0.5,    0.0)
		pg.point(  -0.5,   -0.5,    0.0)
		pg.point(  -0.5,    0.5,    0.0)
		pg.end
	  elsif(c.class == Hash)	# テクスチャ
		pg = PG.new
		glScale(      w,      h,    1.0)
		pg.begin												# 前面
		pg.texture(c)
		pg.texcoord(0.0, 0.0); pg.point(   0.5,    0.5,    0.0)
		pg.texcoord(0.0, 1.0); pg.point(   0.5,   -0.5,    0.0)
		pg.texcoord(1.0, 1.0); pg.point(  -0.5,   -0.5,    0.0)
		pg.texcoord(1.0, 0.0); pg.point(  -0.5,    0.5,    0.0)
		pg.end
	  end
	end

	# 原点を中心に任意の大きさの直方体を出現させる
	def self.cuboid(w, h, d, c)									# 幅, 高さ, 奥行き, 色
		pg = PG.new
		glScale(      w,      h,      d)
		2.times {
			pg.begin											# 後(前)面
			pg.material(*c)
			pg.point(   0.5,    0.5,    0.5)
			pg.point(  -0.5,    0.5,    0.5)
			pg.point(  -0.5,   -0.5,    0.5)
			pg.point(   0.5,   -0.5,    0.5)
			pg.end
			pg.begin											# 上(下)面
			pg.point(   0.5,    0.5,    0.5)
			pg.point(   0.5,    0.5,   -0.5)
			pg.point(  -0.5,    0.5,   -0.5)
			pg.point(  -0.5,    0.5,    0.5)
			pg.end
			glRotatef( 180.0,    1.0,    0.0,    0.0)
		}
		2.times {
			pg.begin											# 左(右)面
			pg.point(   0.5,    0.5,    0.5)
			pg.point(   0.5,   -0.5,    0.5)
			pg.point(   0.5,   -0.5,   -0.5)
			pg.point(   0.5,    0.5,   -0.5)
			pg.end
			glRotatef( 180.0,    0.0,    1.0,    0.0)
		}
	end

	def self.listDraw
		unless(@listID)
			@listID = glGenLists(1)
			puts('%s: %d' % [self.class.to_s, @listID])
			glNewList(@listID, GL_COMPILE)
			yield
			glEndList
		end
		glCallList(@listID)										# 描画
	end
end

#-----------------------------------------------
#
#	プリセットモデル
#
class Preset < Model0

	def self.draw(type, xyz, s, c, opt)
		glPushMatrix
		glTranslatef(  -0.5,   -0.5,   -0.5)
		glTranslatef(*xyz)
		it = opt[:SCALE] and glScale(*it)
		self._draw(type, s, c)
		glPopMatrix
	end

	def self._draw(type, s, c)									# 種別, 大きさ(半径), 色
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, c)
		glScale(      s,      s,      s)
		case(type)
		  when(:SPHERE_S)
			glutSolidSphere(1.0, 12, 9)
		  when(:SPHERE_W)
			glutWireSphere(1.0, 12, 9)
		  when(:TEAPOT_S)
			glScale(    1.0,    1.0,   -1.0)
			glutSolidTeapot(1.0)
		  when(:TEAPOT_W)
			glutWireTeapot(1.0)
		end
	end
end

class Presets < Array

	def draw
		Preset.listDraw {
			self.each {|type, xyz, s, c, opt|
				Preset.draw(type, xyz, s, c, opt)
			}
		}
	end
end

#-----------------------------------------------
#
#	平面
#
class Surface < Model0

	def self.draw(xyz, wh, c)
		glPushMatrix
		glTranslatef(-wh[0] / 2.0, -wh[1] / 2.0, 0.0)
		glTranslatef(*xyz)
		self._draw(*wh, c)
		glPopMatrix
	end

	def self._draw(w, h, c)										# 幅, 高さ, 色
		surface(w, h, c)
	end
end

class Surfaces < Array

	def draw
		Surface.listDraw {
			self.each {|xyz, wh, c|
				Surface.draw(xyz, wh, c)
			}
		}
	end
end

#-----------------------------------------------
#
#	直方体
#
class Cuboid < Model0

	def self.draw(xyz, whd, c)
		glPushMatrix
		glTranslatef(-whd[0] / 2.0, -whd[1] / 2.0, -whd[2] / 2.0)
		glTranslatef(*xyz)
		self._draw(*whd, c)
		glPopMatrix
	end

	def self._draw(w, h, d, c)									# 幅, 高さ, 奥行き, 色
		cuboid(w, h, d, c)
	end
end

class Cuboids < Array

	def draw
		Cuboid.listDraw {
			self.each {|xyz, whd, c|
				Cuboid.draw(xyz, whd, c)
			}
		}
	end
end

#-----------------------------------------------
#
#	線分
#
class Line < Model0

	def self.draw(xyz0, lrd, c, w)
		glPushMatrix
		glTranslatef(  -0.5,   -0.5, -lrd[2] / 2.0)
		glTranslatef(*xyz0)
		self._draw(*lrd, c, w)
		glPopMatrix
	end

	def self._draw(l, r, d, c, w)								# 長さ、角度, 奥行き, 色, 太さ
		glRotatef(r * 180 / PI, 0.0, 0.0, 1.0)
		cuboid(l, w, d, c)
	end
end

class Lines < Array

	def draw
		Line.listDraw {
			self.each {|xyz0, lrd, c, w|
				Line.draw(xyz0, lrd, c, w)
			}
		}
	end
end

#-----------------------------------------------
#
#	ポリゴン
#
class Polygon < Model0

	def self.draw(vertexes, c)
		glPushMatrix
		glTranslatef(  -0.5,   -0.5,   -0.5)
		self._draw(vertexes, c)
		glPopMatrix
	end

	def self._draw(vertexes, c)									# 頂点, 色
		pg = PG.new
		pg.begin
		pg.material(*c)
		vertexes.each {|vertex|
			pg.point(*vertex)
		}
		pg.end
	end
end

class Polygons < Array

	def draw
		Polygon.listDraw {
			self.each {|vertexes, c|
				Polygon.draw(vertexes, c)
			}
		}
	end
end

#-----------------------------------------------
#
#	板状ポリゴン
#
class Polygon2d < Model0

	def self.draw(vertexes, z, d, c)
		glPushMatrix
		glTranslatef(  -0.5,   -0.5,   -0.5)
		self._draw(vertexes, z, z - d, c)
		glPopMatrix
	end

	def self._draw(vertexes, z0, z1, c)							# 頂点, 厚み, 色
		pg = PG.new
		pg.begin
		pg.material(*c)
		vertexes.each {|vertex|									# 表面
			pg.point(*vertex, z1)
		}
		pg.end
		(vs = vertexes.size).times {|n|							# 側面
			pg.begin
			pg.point(*vertexes[n], z1)
			pg.point(*vertexes[n], z0)
			pg.point(*vertexes[(n + 1) % vs], z0)
			pg.point(*vertexes[(n + 1) % vs], z1)
			pg.end
		}
		pg.begin
		pg.material(*c)
		vertexes.reverse.each {|vertex|							# 裏面
			pg.point(*vertex, z0)
		}
		pg.end
	end
end

class Polygon2ds < Array

	def draw
		Polygon2d.listDraw {
			self.each {|vertexes, z, d, c|
				Polygon2d.draw(vertexes, z, d, c)
			}
		}
	end
end

#-----------------------------------------------
#
#	回転体
#
class Revolution < Model0

	@@types = {}

	def self.draw(xyz, rxyz, sxyz, rs, rd, rn, type, csp, c, opt = {})
		glPushMatrix
		glTranslatef(  -0.5,   -0.5,   -0.5)
		glTranslatef(*xyz)
		glRotatef(*rxyz)
		glScale(*sxyz)
		(it = opt[:CB]) and it.call
		self._draw(rs, rd, rn, type, csp, c)
		glPopMatrix
	end

	def self._draw(rs, rd, rn, type, csp, c)					# 開始角, 角増分, 回数, 断面形状, 断面形状パラメータ, 色
		pg = PG.new
		pxyz = [[], []]
		rr = rs; (0..rn).each {|n|								# 水平方向に回転
			cb = n % 2; ab = (n + 1) % 2						# current bank, another bank
			nidx = @@types[type].call(csp) {|idx, ry|			# 順に断面形状を受け取る
				pxyz[cb][idx + 1] = []
				pxyz[cb][idx + 1] << ry[0] * cos(rad(rr))		# x
				pxyz[cb][idx + 1] << ry[1]						# y
				pxyz[cb][idx + 1] << ry[0] * sin(rad(rr))		# z
				if(pxyz[ab][idx] && pxyz[cb][idx])
					pg.begin
					pg.material(*c)
					pg.point(*pxyz[cb][idx    ])
					pg.point(*pxyz[cb][idx + 1])
					pg.point(*pxyz[ab][idx + 1])
					pg.point(*pxyz[ab][idx    ])
					pg.end
				end
			}
			if(nidx && (n == 0 || n == rn))						# 切り口を描画
				pg.begin
				pg.material(*c)
				(1...nidx).each {|idx|
#					pg.material(1.0, 0, 0)
					n ==  0 and pg.point(*pxyz[cb][idx])
					n == rn and pg.point(*pxyz[cb][nidx - idx])
				}
				pg.end
			end
			rr += rd
		}
	end

	#-----------------------------------------------
	#
	#	盤の断面を定義
	#
	#		:R	=> 10.0,										# 半径
	#		:RI	=>  5.0,										# 内径
	#		:H	=> 10.0,										# 高さ
	#
	@@types[:DISC] = Proc.new {|csp, &realize|

		ri = csp[:RI] || 0.001; h = csp[:H] || 0
		realize.call(		0, [csp[:R],	h])
		realize.call(		1, [ri,			h])					# 上から下
		h == 0 and next
		realize.call(	2, [ri,			0])
		realize.call(	3, [csp[:R],	0])
		realize.call(	4, [csp[:R],	h])
		5														# 切り口を描画するなら断面の頂点数を返す
	}

	#-----------------------------------------------
	#
	#	錐の断面を定義
	#
	#		:R	=> 10.0,										# 半径
	#		:H	=> 10.0,										# 高さ
	#
	@@types[:CONE] = Proc.new {|csp, &realize|

		realize.call(0, [0, csp[:H]])
		realize.call(1, [csp[:R], 0])
	}

	#-----------------------------------------------
	#
	#	球の断面を定義
	#
	#		:R	=> 10.0,										# 半径
	#		:RS	=> 0,											# 開始角
	#		:RD	=> 10,											# 角増分
	#		:RN	=> 18,											# 回数
	#
	@@types[:SPHERE] = Proc.new {|csp, &realize|

		csp[:RN].times {|idx|
			rz = csp[:RS] + csp[:RD] * idx						# 真上:0 -> 真下:180
			x = sin(rad(rz)) * csp[:R]
			y = cos(rad(rz)) * csp[:R]
			realize.call(idx, [x, y])
		}
		csp[:RN]
	}

	#-----------------------------------------------
	#
	#	円環の断面を定義
	#
	#		:R	=> 20.0,										# 円環半径
	#		:RT	=>  5.0,										# 環径
	#		:RS	=> 0,											# 開始角
	#		:RD	=> 10,											# 角増分
	#		:RN	=> 36,											# 回数
	#
	@@types[:TORUS] = Proc.new {|csp, &realize|

		csp[:RN].times {|idx|
			rz = csp[:RS] + csp[:RD] * idx						# 内側:0 -> 外側:180 -> 内側: 360
			x = csp[:R] - cos(rad(rz)) * csp[:RT]
			y = sin(rad(rz)) * csp[:RT]
			realize.call(idx, [x, y])
		}
		csp[:RN]
	}
end

class Revolutions < Array

	def draw
		Revolution.listDraw {
			self.each {|xyz, rxyz, sxyz, rs, rd, rn, type, csp, c, opt|
				Revolution.draw(xyz, rxyz, sxyz, rs, rd, rn, type, csp, c, opt)
			}
		}
	end
end

#-----------------------------------------------
#
#	文字
#
class Text < Model0

	def self.draw(cuboids)										# 立方体群, 色
		cuboids.each {|cuboid|
			Cuboid.draw(*cuboid)
		}
	end
end

#===============================================================================
#
#	Class: LegacyGraphics
#
class LegacyGraphics < TrueLegacyGraphicsCommon

	attr_reader :inputs, :sprites, :cur_d

	@@color = []
	(0..7).each {|c|											# イロイッカイズツ
		grb = '%03b' % c; @@color << [grb[1].to_f, grb[0].to_f, grb[2].to_f]
	}
	def color(c)
		_color(c, @@color)
	end

	def self.get_image_size_png(file)
		@@png_image = Magick::Image.read(@@cur_png = file)
		[@@png_image[0].columns, @@png_image[0].rows]
	end

	def self.point(x, y)
		@@png_image[0].pixel_color(x, y)
	end

	def initialize(h_dmy, w_dmy, y_dmy, x_dmy, gy = 400, gx = 640, cy = 16, cx = 8, c = 0, out = {})

		@gx = gx;	@gy = gy
		$jiffies = 0

		glutInit
		glutInitWindowSize(gx, gy)
		glutCreateWindow('TrueLegacyGraphicsOpenGL')
		glutInitDisplayMode(GLUT_RGBA)

		glutReshapeFunc(method(:reshape).to_proc)
		glutDisplayFunc(method(:display).to_proc)

		glEnable(GL_NORMALIZE)
		glEnable(GL_DEPTH_TEST)
		glEnable(GL_CULL_FACE)
#		glEnable(GL_BLEND)
#		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

		glClearColor(*color(c), 1.0)

		@lights_enable = false; @lights = {
			GL_POSITION	=> [ -100,  700, -200,    1 ],			# 0: 無限遠
			GL_DIFFUSE	=> [  0.8,  0.8,  0.8,  1.0 ],
			GL_AMBIENT	=> [  0.5,  0.5,  0.5,  1.0 ],
		}

		@perspective = [ 55.0, gx / gy.to_f, 100.0, 2000.0 ]	# 28mm:65, 35mm:55, 50mm:40

		@lookat = [	    0.0,    0.0, -350.0,					# 視点の位置
					    0.0,    0.0,    0.0,					# 着目点の中心位置
					    0.0,    1.0,    0.0 ]					# 着目点の上ベクトル

		@camera = [	  -30.0,  -40.0,  100.0,					# 視点移動(全体移動) 右, 下, 手前が正
					   15.0,    0.0,    1.0,    0.0 ]

		@arotate = [    0.0,    0.0,    1.0,    0.0 ]			# 物体の自動的な回転

		@inputs = { :LEVEL	=> {}, :EDGE	=> [] }
		glutIgnoreKeyRepeat(1)									# キー入力
		glutSpecialFunc(method(:spkeydown).to_proc)
		glutSpecialUpFunc(method(:spkeyup).to_proc)
		glutKeyboardFunc(method(:keydown).to_proc)
		glutKeyboardUpFunc(method(:keyup).to_proc)

		@callback = method(:callback).to_proc
		@callback.call(nil, [:RESET] + @camera, @cb_params = nil)

		@ox = gx / 2; @oy = gy / 2
		@presets = Presets.new
		@surfaces = Surfaces.new
		@cuboids = Cuboids.new
		@lines = Lines.new
		@polygons = Polygons.new
		@polygon2ds = Polygon2ds.new
		@revolutions = Revolutions.new

		depth3d													# 奥行き位置、奥行き長さ(手前側が正)
	end

	def callback(win, info, params)
		case(info[0])

		  when(:RESET)
			@cx, @cy, @cz, @cry = *info[1, 4]

		  when(:VSYNC)
			while(it = win.inputs[:EDGE].shift)
				puts('edge code = %s' % it.inspect)
			end
			win.inputs[:LEVEL][100] and @cx  -= 4				# ←
			win.inputs[:LEVEL][102] and @cx  += 4				# →
			win.inputs[:LEVEL][101] and @cy  -= 4				# ↑
			win.inputs[:LEVEL][103] and @cy  += 4				# ↓
			win.inputs[:LEVEL]["'"] and @cz  -= 4				# '(前進)
			win.inputs[:LEVEL]['/'] and @cz  += 4				# /(後退)
			win.inputs[:LEVEL][','] and @cry -= 1				# <(取舵)
			win.inputs[:LEVEL]['.'] and @cry += 1				# >(面舵)
			@camera = [ @cx, @cy, @cz, @cry, 0.0, 1.0, 0.0 ]

		  when(:SPKEYDOWN)

		  when(:SPKEYUP)

		  when(:KEYDOWN)
			info[1] == 'r' and @ar = win.arotate3d(@ar ? 0 :  1)	# 自動回転 ON/OFF
			info[1] == 'R' and @ar = win.arotate3d(@ar ? 0 : -1)
			info[1] == 'q' and exit

		  when(:KEYUP)

		end
	end

	def light3d(key, values)
		@lights[key] = values
	end

	def perspective3d(*values)
		@perspective = values
		@perspective[1] ||= @ox / @oy.to_f
	end

	def lookat3d(*values)
		@lookat = values
		@lookat.size == 3 and @lookat += [ 0.0, 0.0, 0.0 ]
		@lookat.size == 6 and @lookat += [ 0.0, 1.0, 0.0 ]
	end

	def camera3d(*values)
		@camera = values
		@camera.size == 3 and @camera += [ 0.0, 0.0, 1.0, 0.0 ]
		@callback.call(nil, [:RESET] + @camera, @cb_params)
	end

	def arotate3d(*values)										# 物体の自動的な回転(速度)
		@arotate_r ||= 0.0
		if(values.size == 0)
			glRotatef(@arotate_r, *@arotate[1, 3])
			@arotate_r += @arotate[0]
		else
			@arotate = values
			@arotate.size == 1 and @arotate += [ 0.0, 1.0, 0.0 ]
		end
		@arotate[0] != 0.0
	end

	def reshape(width, height)
		unless(@lights_enable)
			@lights.each {|key, values|
				glLightfv(GL_LIGHT0, key, values)
			}
			glEnable(GL_LIGHT0)
			glEnable(GL_LIGHTING)
			@lights_enable = true
		end
		glViewport(0, 0, width, height)
		glLoadIdentity
		gluPerspective(*@perspective)
		gluLookAt(*@lookat)
	end

	def display
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

		glPushMatrix
		glRotatef(*@camera[3, 4])								# 視点の向き
		glTranslatef(*@camera[0, 3])							# 視点移動(全体移動)

		arotate3d												# 物体の自動的な回転

		@presets.draw
		@surfaces.draw
		@cuboids.draw
		@lines.draw
		@polygons.draw
		@polygon2ds.draw
		@revolutions.draw

		it = @sprites and it.draw
		glPopMatrix

		glFlush
	end

	def depth3d(z = 0, d = 1)									# 手前が正
		@cur_z = -z; @cur_d = d
	end

	def sphere3d(pos, s, c, opt = {})							# pos = [x, y, z], s = size
		@presets << __sphere3d(pos, s, c, opt)
	end
	def __sphere3d(pos, s, c, opt)
		[ opt[:WIRE] ? :SPHERE_W : :SPHERE_S,
			[ @ox - pos[0], @oy - pos[1], -pos[2] ], s, color(c), opt ]
	end

	def teapot3d(pos, s, c, opt = {})							# pos = [x, y, z], s = size
		@presets << __teapot3d(pos, s, c, opt)
	end
	def __teapot3d(pos, s, c, opt)
		[ opt[:WIRE] ? :TEAPOT_W : :TEAPOT_S,
			[ @ox - pos[0], @oy - pos[1], -pos[2] ], s, color(c), opt ]
	end

	def polygon3d(vertexes, c)									# vertexes = [ [x1, y1, z1], [x2, y2, z2], [x3, y3, z3]...]
		@polygons << __polygon3d(vertexes, c)
	end
	def __polygon3d(vertexes, c)
		vs = []; vertexes.each {|x, y, z|
			vs << [	@ox - x,	@oy - y,	-z ]
		}
		[	vs,	color(c) ]
	end

	def fill(x1, y1, x2, y2, c)
		if(@cur_d == 0)
			@surfaces << __surface(x1, y1, x2, y2, c)			# 平面の面
		else
			@cuboids << __cuboid(x1, y1, x2, y2, c)				# 立体の面
		end
	end
	def __surface(x1, y1, x2, y2, c)
		[	[	@ox - x1,		@oy - y1,		@cur_z ],
			[	x2 - x1 + 1,	y2 - y1 + 1 ],				color(c) ]
	end
	def __cuboid(x1, y1, x2, y2, c)
		[	[	@ox - x1,		@oy - y1,		@cur_z ],
			[	x2 - x1 + 1,	y2 - y1 + 1,	@cur_d ],	color(c) ]
	end

	def _line(x1, y1, x2, y2, c, w, d, a)
		@lines << __line(x1, y1, x2, y2, c, w, d, a)
	end
	def __line(x1, y1, x2, y2, c, w, d, a)
		[	[	@ox - (x1 + x2) / 2.0,	@oy - (y1 + y2) / 2.0,	@cur_z ],
			[	sqrt(((x1 - x2) ** 2 + (y1 - y2) ** 2).to_f) + 1,
				atan((y2 - y1) / (x2 - x1).to_f),	@cur_d ],	color(c),	w ]
	end

	def polygon(points, c, f = '', w = 1)						# vertexes = [ [x1, y1], [x2, y2], [x3, y3]...]
		@polygon2ds << __polygon(points, c, f, w)
	end
	def __polygon(points, c, f, w)
		vs = []; points.each {|x, y|
			vs << [	@ox - x,	@oy - y ]
		}
		[	vs,	@cur_z,	@cur_d,	color(c) ]
	end

	def cone(x, y, r, h, c, opt = {})
		@revolutions << __cone(x, y, r, h, c, opt)
	end
	def __cone(x, y, r, h, c, opt)
		rn = opt[:DIV] || 36; rd = 360.0 / rn; rs = rd / 2
		[															# 角錐(円錐)
			[	@ox - x,		@oy - y,		@cur_z ],			# 位置
			opt[:DIR] || [ -90.0,    1.0,    0.0,    0.0 ],			# 角度
			[   1.0,    1.0,    1.0 ],								# サイズ(比率)
			rs, rd, rn, :CONE, {
				:R	=> r,											# 半径
				:H	=> h,											# 高さ
			}, color(c), opt
		]
	end

	def circle(x, y, r, c, f = '', dy = 1, opt = {})
		@revolutions << __circle(x, y, r, c, f, dy, opt)
	end
	def __circle(x, y, r, c, f, dy, opt)
		sa = deg(opt[:SANGLE] || 0); fa = deg(opt[:FANGLE] || 2 * PI); da = sa - fa
		[															# 平面/立体の輪(管)、円盤
			[	@ox - x,		@oy - y,		@cur_z ],			# 位置
			opt[:DIR] || [ -90.0,    1.0,    0.0,    0.0 ],			# 角度
			[   1.0,    1.0,    opt[:RATIO] || 1.0 ],				# サイズ(比率(適当))
			180 - sa, da / 36.0, 36, :DISC, {
				:R	=> r + 0.5,										# 半径
				:RI	=> f =~ /f/ ? nil : r - (opt[:W] || 1) + 0.5,	# 内径
				:H	=> @cur_d,										# 管長
			}, color(c), opt
		]
	end

	def put_image_png(x, y, file, opt = {})
		@textures ||= {}
		unless(texture = @textures[file])
			w, h = LegacyGraphics.get_image_size_png(file); pixels = ''
			h.times {|y|
				w.times {|x|
					c = LegacyGraphics.point(x, y)
					pixels << [c.red >> 8, c.green >> 8, c.blue >> 8, c.alpha >> 8].pack('C*')
#					pixels << [c.red >> 8, c.green >> 8, c.blue >> 8, 128].pack('C*')
				}
			}
			texture = @textures[file] = {
				:W		=> w,
				:H		=> h,
				:PIXELS	=> pixels,
			}
		end
		@surfaces << __surface_with_texture(x, y, x + texture[:W], y + texture[:H], texture)
	end
	def __surface_with_texture(x1, y1, x2, y2, texture)
		[	[	@ox - x1,		@oy - y1,		@cur_z ],
			[	x2 - x1 + 1,	y2 - y1 + 1 ],				texture ]
	end

	def _text(x, y, s, c, ps = {})
	end
	def __text(x, y, s, c, ps)
		ps[:PSET_FUNC] = method(:__cuboid)
		cuboids = []; self.__text_bmf(x, y, s, c, ps) {|cuboid|
			cuboids << cuboid
		}
		[cuboids]
	end

	def setcallback3d(callback, params = nil)
		@callback = callback
		@cb_params = params
	end

	#---------------------------------------------------------------
	#
	#	各種入力
	#
	def spkeydown(key, x, y)
		@inputs[:EDGE] << key
		@inputs[:LEVEL][key] = true
		@callback.call(self, [:SPKEYDOWN, key], @cb_params)
	end

	def spkeyup(key, x, y)
		@inputs[:LEVEL][key] = false
		@callback.call(self, [:SPKEYUP, key], @cb_params)
	end

	def keydown(key, x, y)
		@inputs[:EDGE] << key
		@inputs[:LEVEL][key] = true
		@callback.call(self, [:KEYDOWN, key], @cb_params)
	end

	def keyup(key0, x, y)
		key = key0.is_a?(Integer) ? key0.chr : key0					# BUG?
		@inputs[:LEVEL][key] = false
		@callback.call(self, [:KEYUP, key], @cb_params)
	end

	#---------------------------------------------------------------
	#
	#	垂直帰線割り込みのシミュレート
	#
	def vsync(dummy)
		@spf = (1.0 / @fps)											# フレーム速度(1/FPS)

		$jiffies += 1
		@start = Time.now.to_f; @next = @start + @spf

		@callback.call(self, [:VSYNC], @cb_params)					# コールバック処理
		glutPostRedisplay											# 描画処理

		@finish = Time.now.to_f
		puts('now: %13.2f, load: %6.2f%%' % [@start, (@finish - @start) * 100 / @spf]) if($jiffies % 30 == 0)
		glutTimerFunc((it = (@next - @finish) * 1000) > 0 ? it : 1, method(:vsync).to_proc, 0)
	end

	def refresh(fps = 30)

		# 周期処理(vsync)を開始
		@fps = fps and glutTimerFunc(1, method(:vsync).to_proc, 0)

		glutMainLoop
	end

	#---------------------------------------------------------------
	#
	#	表示内容を画像として保存
	#		https://github.com/lazyatom/ruby-opengl/blob/master/examples/misc/readpixel.rb
	#
	def write_to_file(file)
		pixels = glReadPixels(0, 0, @gx, @gy, GL_RGBA, GL_UNSIGNED_SHORT)
		image = Magick::Image.new(@gx, @gy)
		image.import_pixels(0, 0, @gx, @gy, 'RGBA', pixels, Magick::ShortPixel)
		image.flip!
		image.write(file)
	end
end

class Sprite

	def initialize(win)
		@win = win
	end

	def sphere3d(pos, s, c, opt = {})
		@draw = Preset.method(:draw)
		@data = @win.__sphere3d(pos, s, c, opt)
	end

	def teapot3d(pos, s, c, opt = {})
		@draw = Preset.method(:draw)
		@data = @win.__teapot3d(pos, s, c, opt)
	end

	def polygon3d(vertexes, c)
		@draw = Polygon.method(:draw)
		@data = @win.__polygon3d(vertexes, c)
	end

	def fill(x1, y1, x2, y2, c)
		if(@win.cur_d == 0)
			@draw = Surface.method(:draw)
			@data = @win.__surface(x1, y1, x2, y2, c)
		else
			@draw = Cuboid.method(:draw)
			@data = @win.__cuboid(x1, y1, x2, y2, c)
		end
	end

	def line(x1, y1, x2, y2, c, bf = '', w = 1, d = nil, a = nil)
		@draw = Line.method(:draw)
		@data = @win.__line(x1, y1, x2, y2, c, w, d, a)
	end

	def polygon(vertexes, c, f = '', w = 1)
		@draw = Polygon2d.method(:draw)
		@data = @win.__polygon(vertexes, c, f, w)
	end

	def cone(x, y, r, h, c, opt = {})
		@draw = Revolution.method(:draw)
		@data = @win.__cone(x, y, r, h, c, opt)
	end

	def circle(x, y, r, c, f = '', dy = 1, opt = {})
		@draw = Revolution.method(:draw)
		@data = @win.__circle(x, y, r, c, f, dy, opt)
	end

	def text(x, y, s, c, ps = {})
		@draw = Text.method(:draw)
		@data = @win.__text(x, y, s, c, ps)
	end

	def draw
		@draw and @draw.call(*@data)
	end
end

class Sprites < Hash

	def draw
		self.each {|key, sprite|
			sprite.draw
		}
	end
end

__END__

