#!/usr/bin/env ruby

require './TrueLegacyGraphicsCairo'

require 'optparse'
opt_size = opt_origin = opt_step = opt_remove_inner = opt_join_inner = opt_one_stroke = opt_show_tips = opt_progress = nil
opts = OptionParser.new
opts.version = '1.10'; opts.release = 'a'
opts.banner = "Usage: ./cutDotPat.rb [options] [imagefile]"
opts.on('-s', '--size x,y',						'sample size.')				{|p| opt_size			= p }
opts.on('-O', '--origin x,y',					'sample origin.')			{|p| opt_origin			= p }
opts.on('-S', '--step x,y',						'sample step.')				{|p| opt_step			= p }
opts.on('-l', '--remove-inner',					'remove inner lines.')		{|p| opt_remove_inner	= true }
opts.on('-j', '--join-inner',					'join inner lines.')		{|p| opt_join_inner		= true }
opts.on('-o', '--one-stroke',					'one stroke.')				{|p| opt_one_stroke		= true }
opts.on('-t', '--show-tips',					'show one stroke tips.')	{|p| opt_show_tips		= true }
opts.on('-a', '--progress',						'make progress images.')	{|p| opt_progress		= true }
params = opts.permute(ARGV)

#win = LegacyGraphics.new(0, 0, 0, 0,  838,  844, 16, 8, false, { :file => 'cutdotpat', :type => 'svg' })	# png/pdf/svg/ps
win = LegacyGraphics.new(0, 0, 0, 0,  838,  844, 16, 8, 7, { :file => 'cutdotpat', :type => 'png' })	# png/pdf/svg/ps

pat = [
	[9, 9, 9, 9, 9, 9, 9, 9],
	[9, 7, 9, 9, 9, 9, 9, 9],
	[9, 7, 6, 6, 9, 9, 9, 9],
	[9, 7, 6, 6, 6, 6, 9, 9],
	[9, 7, 6, 6, 6, 6, 6, 6],
	[9, 7, 6, 6, 6, 6, 9, 9],
	[9, 7, 6, 6, 9, 9, 9, 9],
	[9, 7, 9, 9, 9, 9, 9, 9],
	[9, 7, 9, 9, 9, 9, 9, 9],
	[2, 2, 2, 9, 9, 9, 9, 9],
	[2, 2, 2, 9, 9, 9, 9, 9],
	[9, 9, 9, 9, 9, 9, 9, 9],
]

class String
	def to_ia
		r = []; self.split(/[,x+:]/).each {|ns|
			r << ns.to_i
		}
		r
	end
end

if(file = params[0])
	origin	= (it = opt_origin)	? it.to_ia : [0, 0]
	size	= (it = opt_size)	? it.to_ia : LegacyGraphics.get_image_size_png(file)
	step	= (it = opt_step)	? it.to_ia : [1, 1]
	pat = win.get(*origin, *size, file, *step) {|c|
		c.opacity > 0 ? 9 : (c.green > 0x7FFF ? 4 : 0) + (c.red > 0x7FFF ? 2 : 0) + (c.blue > 0x7FFF ? 1 : 0)
	}
end

class Grid < Hash

	def initialize(d)
		@dx = (d == :H) ? 1 : 0
		@dy = (d == :H) ? 0 : 1
	end

	def set(x, y, c = 7)						# grid['x, y'] = [lap, x, y, c, len], lap(2: 外形, 3: 色境界, 4: 同色)
		self[key = '%d, %d' % [x, y]] ? self[key][0] += (self[key][3] == c ? 2 : 1) : self[key] = [2, x, y, c, 1]
	end

	def remove_laps(remove_inner = false)
		self.each {|k, v|
			v[0] > (remove_inner ? 2 : 3) and self.delete(k)
		}
	end

	def join_lines(join_inner = false)
		self.keys.reverse.each {|k|
			cur = self[k]
			prv = self['%d, %d' % [cur[1] - @dx, cur[2] - @dy]]
			if(prv and (join_inner or cur[0] == prv[0]))
				prv[4] += cur[4]
				self.delete(k)
			end
		}
	end

	def line_each
		self.each {|xy, attr|
			yield(attr.object_id, attr[1], attr[2], attr[1] + attr[4] * @dx, attr[2] + attr[4] * @dy)
		}
	end
end

class OneStroker

	attr_reader :tips

	# @xys['x1, y1'] = ['LINExx', 'LINEyy']
	# @xys['x2, y2'] = ['LINExx', 'LINEzz']
	# @lines['LINExx, x1, y1'] = ['LINExx', x2, y2]

	def initialize(hgrid, vgrid)
		@xys = {}; @lines = {}
		hgrid.line_each {|line_id, x1, y1, x2, y2|
			@xys[xy1 = '%s, %s' % [x1, y1]] ||= []; 				@xys[xy2 = '%s, %s' % [x2, y2]] ||= []
			@xys[xy1] << line_id;									@xys[xy2] << line_id
			@lines['%s, %s' % [line_id, xy1]] = [line_id, x2, y2];	@lines['%s, %s' % [line_id, xy2]] = [line_id, x1, y1];	
		}
		vgrid.line_each {|line_id, x1, y1, x2, y2|
			@xys[xy1 = '%s, %s' % [x1, y1]] ||= []; 				@xys[xy2 = '%s, %s' % [x2, y2]] ||= []
			@xys[xy1] << line_id;									@xys[xy2] << line_id
			@lines['%s, %s' % [line_id, xy1]] = [line_id, x2, y2];	@lines['%s, %s' % [line_id, xy2]] = [line_id, x1, y1];	
		}
		@tips = []; @xys.keys.sort.each {|xy|	# 奇数点を抽出する
			@xys[xy].size % 2 != 0 and @tips << '%s, %s' % [@xys[xy][0], xy]
		}
	end

	def each
		line2another = nil; loop {

			@lines.size == 0 and break

			unless(@lines[line2another])
				line2another = nil; while(tip = @tips.pop)
					@lines[tip] and line2another = tip and break
				end
				line2another ||= @lines.keys[0]
			end

			lxy2 = @lines[line2another]; cur_line = lxy2[0]; ant_x = lxy2[1]; ant_y = lxy2[2]
			lxy1 = @lines[line2current = '%d, %s' % [cur_line, ant_xy = '%d, %d' % [ant_x, ant_y]]]; cur_x = lxy1[1]; cur_y = lxy1[2]

			yield(cur_line, cur_x, cur_y, ant_x, ant_y)
			@lines.delete(line2another)
			@lines.delete(line2current)

			line2another = @xys[ant_xy].each {|line_id|
				@lines['%d, %s' % [line_id, ant_xy]] and break('%d, %s' % [line_id, ant_xy])
			}
		}
	end
end

hgrid = Grid.new(:H); vgrid = Grid.new(:V)

# すべてのドットを四角に線で囲む
ry = 0; pat.each {|hpat|
	rx = 0; hpat.each {|c|
		if(c < 8)
			hgrid.set(rx,     ry,     c)
			hgrid.set(rx,     ry + 1, c)
			vgrid.set(rx,     ry,     c)
			vgrid.set(rx + 1, ry,     c)
		end
		rx += 1
	}
	ry += 1
}

hgrid.remove_laps(opt_remove_inner);	vgrid.remove_laps(opt_remove_inner)	# 重なった線を削除する(内部の線も削除する)
hgrid.join_lines(opt_join_inner);		vgrid.join_lines(opt_join_inner)	# 隣り合う線をつなげる(内部の線もつなげる)

class Numeric
	def to_in
		self * 72 / 25.4
	end
end

unless(opt_one_stroke)
	# 横線、縦線の順に描画
	draw = Proc.new {|win, hgrid, vgrid, x0, y0, t, &callback|
		callback.call(n = 0)
		hgrid.each {|k, v|
			win.line(xx = x0 + v[1] * t, yy = y0 + v[2] * t, xx + v[4] * t, yy, v[0] == 2 ? 0 : 2)
			callback.call(n += 1)
		}
		vgrid.each {|k, v|
			win.line(xx = x0 + v[1] * t, yy = y0 + v[2] * t, xx, yy + v[4] * t, v[0] == 2 ? 0 : 2)
			callback.call(n += 1)
		}
	}
else
	# 一筆書きで描画
	draw = Proc.new {|win, hgrid, vgrid, x0, y0, t, &callback|
		ost = OneStroker.new(hgrid, vgrid)
		ost.tips.each {|tip|
			idxy = tip.split(/,/); x1 = idxy[1].to_i; y1 = idxy[2].to_i
			win.circle(x0 + x1 * t, y0 + y1 * t, 2, 1)
		} if(opt_show_tips)
		callback.call(n = 0)
		ost.each {|c, x1, y1, x2, y2|
			win.line(x0 + x1 * t, y0 + y1 * t, x0 + x2 * t, y0 + y2 * t, 0)
			callback.call(n += 1)
		}
	}
end

# 図化する
#x0 = 8.0.to_in; y0 = 8.0.to_in; t = 4.0.to_in
x0 = 6.0.to_in; y0 = 6.0.to_in; t = 3.0.to_in
x0 = 4.0.to_in; y0 = 4.0.to_in; t = 2.0.to_in
draw.call(win, hgrid, vgrid, x0, y0, t) {}
win.refresh
win.close

# 描画経過を図化する
if(opt_progress)
#	win = AncientGraphics.new(0, 0, 0, 0,  838,  844, 16, 8, 7, { :file => 'cutdotpat', :type => 'png' }) and Dir.mkdir('cutdotpat.d') rescue true
	win = LegacyGraphics.new(0, 0, 0, 0,  226,  226, 16, 8, 7, { :file => 'cutdotpat', :type => 'png' }) and Dir.mkdir('cutdotpat.d') rescue true
	draw.call(win, hgrid, vgrid, x0, y0, t) {|n|
		win.refresh and win.write('cutdotpat.d/%04d' % [n + 100])
	}
	win.close
	# convert -delay 20 -loop 0 -adjoin cutdotpat.d/*.png anime.gif
end

__END__

