#!/usr/bin/env ruby

require 'bundler/setup'

require './TrueLegacyGraphicsCairo'

def koch_curve(cvs, n, l)
	if(n == 0)
		cvs.tForward(l)
	else
		n -= 1; l /= 3.0
		koch_curve(cvs, n, l)
		cvs.tLeft(60)
		koch_curve(cvs, n, l)
		cvs.tRight(120)
		koch_curve(cvs, n, l)
		cvs.tLeft(60)
		koch_curve(cvs, n, l)
	end
end

def koch_curve(cvs, n, l, r = 60)				# for ANIME
	if(n == 0)
		cvs.tForward(l)
	else
		n -= 1; l /= 3.0
		c = cos(rad(r)) * l
		koch_curve(cvs, n, l * 1.5 - c, r)
		cvs.tLeft(r)
		koch_curve(cvs, n, l, r)
		cvs.tRight(r * 2)
		koch_curve(cvs, n, l, r)
		cvs.tLeft(r)
		koch_curve(cvs, n, l * 1.5 - c, r)
	end
end

def c_curve(cvs, n, l)
	if(n == 0)
		cvs.tForward(l)
	else
		n -= 1; l /= sqrt(2)
		cvs.tLeft(45)
		c_curve(cvs, n, l)
		cvs.tRight(90)
		c_curve(cvs, n, l)
		cvs.tLeft(45)
	end
end

def r_dragon(cvs, n, l)
	if(n == 0)
		cvs.tForward(l)
	else
		n -= 1; l /= sqrt(2)
		cvs.tLeft(45)
		r_dragon(cvs, n, l)
		cvs.tRight(90)
		l_dragon(cvs, n, l)
		cvs.tLeft(45)
	end
end

def r_dragon(cvs, n, l)							# for ANIME
	if(n == 0)
		@hue ||= 0; cvs.tSetPenColor([:hsv, @hue, 1, 1]); @hue += 0.2
		cvs.tForward(l)
	else
		n -= 1; l /= sqrt(2)
		cvs.tLeft(45)
		r_dragon(cvs, n, l)
		cvs.tRight(90)
		l_dragon(cvs, n, l)
		cvs.tLeft(45)
	end
end

def l_dragon(cvs, n, l)
	if(n == 0)
		cvs.tForward(l)
	else
		n -= 1; l /= sqrt(2)
		cvs.tRight(45)
		r_dragon(cvs, n, l)
		cvs.tLeft(90)
		l_dragon(cvs, n, l)
		cvs.tRight(45)
	end
end

def tree(cvs, n, l)
	cvs.tSetPenWidth(n * 2 + 1)
	if(n == 0)
		cvs.tForward(l)
		cvs.tBack(l)
	else
		cvs.tForward(l)
		cvs.tRight(25)
		tree(cvs, n - 1, l / 1.5)
		cvs.tLeft(55)
		tree(cvs, n - 1, l / 1.3)
		cvs.tRight(30)
		cvs.tBack(l)
	end
end

def tree(cvs, n, l, p = 1, r = 0)				# for ANIME
	cvs.tSetPenWidth(n * 2 + 1)
	if(n == 0)
		cvs.tPushStat
		cvs.tSetPenColor(:dark_green)
		cvs.tForward(lp = l * p)
		cvs.tBack(lp)
		cvs.tPopStat
	else
		cvs.tSetPenColor(:brown)
		cvs.tForward(l)
		cvs.tRight(25 - r)
		tree(cvs, n - 1, l / 1.5, p, r)
		cvs.tLeft(55)
		tree(cvs, n - 1, l / 1.3, p, r)
		cvs.tRight(30 + r)
		cvs.tBack(l)
	end
end

unless(ARGV[0])
	raise(<<USAGE)
Usage: #{$0} [koch|ccurve|dragon|tree]
USAGE
end

def create_tcanvas(name, n, dir = nil, bgc = 7)
	dir and name = '%s/%s' % [dir += '.d', name] and Dir.mkdir(dir) rescue true
	LegacyGraphics.new(0, 0, 0, 0, 512, 768, 16, 8, bgc, { :file => name % n, :type => 'png' }).tHome
end

if(ARGV[0] =~ /k/)
	d = 2; (720 / d).times {|n|
		cvs = create_tcanvas('koch_%05d', n, 'koch')
		cvs.tSetPos(34, 256)
		cvs.tSetHeading(90)
		koch_curve(cvs, 5, 600, n * d)
		cvs.refresh
		cvs.close
	}
#	ffmpeg -r 24 -i koch.d/koch_%5d.png -vcodec libx264 -s 768x512 -r 24 koch.mp4

elsif(ARGV[0] =~ /c/)
	13.times {|n|
		cvs = create_tcanvas('c_curve_%05d', n, 'c')
		cvs.tSetPos(384, 384)
		cvs.tSetHeading(0)
		c_curve(cvs, n, 240)
		cvs.refresh
		cvs.close
	}

elsif(ARGV[0] =~ /d/)
	d = 4; 360.times {|n|
		@hue = 360 * n * d / 360
		cvs = create_tcanvas('dragon_%05d', n, 'dragon', :black)
		cvs.tSetPos(384, 384)
		cvs.tSetHeading(0)
		r_dragon(cvs, 12, 240)
		cvs.refresh
		cvs.close
	}
#	ffmpeg -r 24 -i dragon.d/dragon_%5d.png -vcodec libx264 -s 768x512 -r 24 dragon.mp4

else
	d1 = 20; d2 = 180; u = 45; c = 30
	540.times {|n|
		cvs = create_tcanvas('tree_%05d', n, 'tree', :sky_blue)
		cvs.tSetPos(384, 384)
		cvs.tSetHeading(0)
		if(n < 180)
			tree(cvs, n / d1, 96, n % d1 / d1.to_f)
		elsif(n < 360)
			nc = n + c; r = (it = nc / u) % 2 == 0 ? nc - it * u : u - (nc - it * u)
			tree(cvs, d2 / d1, 96, 1, r - c)
		else
			tree(cvs, d2 / d1, 96, 1, n % 180 * 2)
		end
		cvs.refresh
		cvs.close
	}
#	ffmpeg -r 24 -i tree.d/tree_%5d.png -vcodec libx264 -s 768x512 -r 24 tree.mp4

end

__END__

