#!/usr/bin/env ruby
 
include Math

def deg2rad(deg)
	deg * PI / 180.0
end

def rad2deg(rad)
	rad * 180.0 / PI
end

class Gpggas < Array

	@@earth_r = 6378.137; @@my_lng = 35.12
	@@lat_l = 2 * PI * @@earth_r * cos(deg2rad(@@my_lng)) / 360.0
	@@lng_l = 2 * PI * @@earth_r / 360.0

	attr_reader :x_min, :x_max, :y_min, :y_max, :zl

	def initialize(zl = 15)
		@x_min = 9999999;	@x_max = 0	# 西 -> 東
		@y_min = 9999999;	@y_max = 0	# 北 -> 南
		@zl = zl
	end

	def setopts(opts)
		@opts = opts
	end

	def <<(l)
		g = Gpgga.new(l, @zl)
		if(g.valid)
			it = @opts[:FROM] and g.time < it and return
			it = @opts[:TO]   and g.time > it and return
			self.last and g.time == self.last.time and return					# 同じ時刻の情報は捨てる
			@zl == 14 and g.tx < 14000 and raise(l)
			@x_min > g.tx and @x_min = g.tx - 1
			@x_max < g.tx and @x_max = g.tx + 1
			@y_min > g.ty and @y_min = g.ty - 1
			@y_max < g.ty and @y_max = g.ty + 1
		end
		super(g)
	end

	def wide(x0 = -1, y0 = -1, x9 = 1, y9 = 1)
		@x_min += x0; @x_max += x9
		@y_min += y0; @y_max += y9
	end

	def	cal_vec(g0, g1)
		return(false) unless(g0.valid and g1.valid)
		[	rad2deg(atan2(g1.wcy - g0.wcy, g1.wcx - g0.wcx)),
			sqrt(((g1.lat_d - g0.lat_d).abs * @@lat_l) ** 2 + ((g1.lng_d - g0.lng_d).abs * @@lng_l) ** 2)	]
	end

	def lopix_xy0(tx, ty, gs = nil)
		gs or return([(tx - @x_min) << 8, (ty - @y_min) << 8])
		txy = 2 ** (gs.zl - @zl)
		[(tx * txy - gs.x_min) << 8, (ty * txy - gs.y_min) << 8]
	end

	def lopix_xy(g)
		[((g.tx - @x_min) << 8) + g.lx, ((g.ty - @y_min) << 8) + g.ly]
	end
end

class Gpgga0

	@@basetime = Time.now.getutc

	def initialize(zl = 15)
		@zl = zl
	end

	def self.maketime(str)
		if(str.gsub(/[^+\d]/, '') =~ /^(\+*)(\d{2})(\d{2})(\d{2}?)/)
			ts = @@basetime.to_a; ts[0] = $4 ? $4.to_i : 0; ts[1] = $3.to_i; ts[2] = $2.to_i
			Time.local(*ts) + 86400 * $1.size
		else
			raise('Time format error.')
		end
	end

	def self.norm_tile(tiles)
		xs = []; ys = []
		tiles.each {|tile|
			xs << tile[0]; ys << tile[1]
		}
		return([xs.min, xs.max, ys.min, ys.max])
	end

	def deg2tile(lat_d, lng_d)
		lat_r = deg2rad(lat_d);	lng_r = deg2rad(lng_d)
		wcx = rad2wcx(lat_r);	wcy = rad2wcy(lng_r)
		pcx = wc2pc(wcx);		pcy = wc2pc(wcy)
		pc2tile(pcx, pcy)
	end

	def dddmm2deg(dddmm)
		dddmmf = dddmm.to_f
		deg = (dddmmf / 100).to_i
		deg + (dddmmf - deg * 100) / 60
	end

	def rad2wcx(rad)
		r = 128 / PI
		r * (rad + PI)
	end

	def rad2wcy(rad)
		r = 128 / PI
		-r / 2 * log((1 + sin(rad)) / (1 - sin(rad))) + 128
	end

	def wc2pc(wc, zl = @zl)
		wc * 2 ** zl
	end

	def pc2tile(pcx, pcy)
		[	(pcx / 256).to_i,
			(pcy / 256).to_i,
			 pcx % 256,
			 pcy % 256,
		]
	end
end

class Gpgga < Gpgga0

	@@lasthour = '00'

	attr_reader :valid, :exs, :raws
	attr_reader :time
	attr_reader :lat_d, :lng_d
	attr_reader :lat_r, :lng_r
	attr_reader :wcx, :wcy
	attr_reader :pcx, :pcy
	attr_reader :tx, :ty, :lx, :ly
	attr_reader :sat_n, :accu, :height

	#  1:UTC時刻
	#  2:緯度
	#  3:北緯/南緯
	#  4:経度
	#  5:東経/西経
	#  6:位置特定品質
	#  7:使用衛星数
	#  8:水平精度低下率
	#  9:海抜高さ
	# 10:メートル
	# 11:ジオイド高さ
	# 12:メートル
	# 13:DGPS使用
	# 14:差動基準地点ID*チェックサム

	def initialize(l, zl = 15)
		@valid = false; @exs = {}; @raws = l.split(/,/); @zl = zl
		@raws[6] == '0' and return
		@time = nil; if(@raws[1] =~ /^(\d{2})(\d{2})(\d{2})\./)
			@@lasthour > $1 and @@basetime += 86400
			ts = @@basetime.to_a; ts[0] = $3.to_i; ts[1] = $2.to_i; ts[2] = (@@lasthour = $1).to_i
			@time = Time.utc(*ts).getlocal
		end
		@time or raise('Unexpected time. [%s]' % l)
		@lat_d = dddmm2deg(@raws[4]);	@lng_d = dddmm2deg(@raws[2])
		((@lat_d - 135.0).abs > 8) and raise(@raws[4])
		((@lng_d - 35.0).abs > 7) and raise(@raws[2])
		@lat_r = deg2rad(@lat_d);		@lng_r = deg2rad(@lng_d)
		@wcx = rad2wcx(@lat_r);			@wcy = rad2wcy(@lng_r)
		@pcx = wc2pc(@wcx);				@pcy = wc2pc(@wcy)
		@tx, @ty, @lx, @ly = pc2tile(@pcx, @pcy)
		@sat_n = @raws[7].to_i
		@accu = @raws[8].to_f
		@height = @raws[9].to_f
		@valid = true
	end
end

class Gsensors < Array

	attr_reader :x_min, :x_max, :y_min, :y_max, :z_min, :z_max

	def initialize
		@x_min = [0, nil]; @x_max = [0, nil]
		@y_min = [0, nil]; @y_max = [0, nil]
		@z_min = [0, nil]; @z_max = [0, nil]
	end

	def <<(ls)
		(ls0s = ls[0].split(/,/)).size == 4 and vs = [] and ls0s[1, 3].each {|v| vs << v.to_i }
		@x_min[0] > vs[0] and @x_min = [vs[0], ls[1]]
		@x_max[0] < vs[0] and @x_max = [vs[0], ls[1]]
		@y_min[0] > vs[1] and @y_min = [vs[1], ls[1]]
		@y_max[0] < vs[1] and @y_max = [vs[1], ls[1]]
		@z_min[0] > vs[2] and @z_min = [vs[2], ls[1]]
		@z_max[0] < vs[2] and @z_max = [vs[2], ls[1]]
		super(vs)
	end

	def peaks
		[@x_min, @x_max, @y_min, @y_max, @z_min, @z_max]
	end
end

class GsensorsPerMin < Hash

	def initialize
	end

	def <<(ls)
		unless(self[it = ls[1].time.strftime('%H%M')])
			self[it] = {
				:x_min	=> [0, nil],	:x_max	=> [0, nil],
				:y_min	=> [0, nil],	:y_max	=> [0, nil],
				:z_min	=> [0, nil],	:z_max	=> [0, nil],
			}
		end
		(ls0s = ls[0].split(/,/)).size == 4 and vs = [] and ls0s[1, 3].each {|v| vs << v.to_i }
		self[it][:x_min][0] > vs[0] and self[it][:x_min] = [vs[0], ls[1]]
		self[it][:x_max][0] < vs[0] and self[it][:x_max] = [vs[0], ls[1]]
		self[it][:y_min][0] > vs[1] and self[it][:y_min] = [vs[1], ls[1]]
		self[it][:y_max][0] < vs[1] and self[it][:y_max] = [vs[1], ls[1]]
		self[it][:z_min][0] > vs[2] and self[it][:z_min] = [vs[2], ls[1]]
		self[it][:z_max][0] < vs[2] and self[it][:z_max] = [vs[2], ls[1]]
	end
end

__END__

